// lib3d.h - header file for 3d library
//
// By Jum Hig
// update: 05:48 99/11/13
//
// changed 10/6/1999:
// updated poly defn to handle include a single colour for the polygon,
// and colour values for poly verts (more compatible).
// also, poly type defn (refers to Allegro poly types :)
//

#define E3D_VERSION		16
#define E3D_MODEL_VERSION	3
#define LIGHT_DISTANT	1
#define LIGHT_OMNI		2

#define MAX_MODELS		100
#define MAX_INSTS		1000

#define MAX_VERTS_PER_MODEL	512
#define MAX_POLYS_PER_MODEL	512

#define MAX_VERTS_PER_LANDSCAPE		2048
#define MAX_POLYS_PER_LANDSCAPE		2048

#define RAD2DEG			57.29578

// Instance flags
#define ADD_TO_SURFLIST		1



// Set up structs
typedef struct {
	float x, y, z;
} Vector_f;

typedef struct {
	float x, y;
} Pt_f;

typedef struct {
	float r, g, b;
} rgb_f;

typedef struct {
	float x, y, z;
	float r, g, b;
} Vertex_f;

typedef struct {
	short type, numverts;
	short p[6];							// vertex indices
	short n[6];							// normal indices
	rgb_f colour;						// polygon colour
	rgb_f vcol[6];						// vertex colours
	short shade;		// 0 = no shade, 1 = shade
	short gouraud;		// 0 = flat, 1 = gouraud
	short filling;		// 0 = solid colour, 1 = colour grad, 2 = texture
	short texmap;		// 1+ = texmap index
	Pt_f texp[6];
	float dist;
} Pgon;


typedef struct {
	short numverts;
	Vertex_f vert[12];
    Pt_f texp[12];
} ClippedPgon;
 
typedef struct {
	Vertex_f pnt[12];
	short numverts;
    Pt_f texp[12];
	float dist;
} MappedPolygon;

typedef struct {
	Vertex_f pnt[12];
	Pt_f texp[12];
} ClippedMappedPolygon;

typedef struct {
	int h, w;
	unsigned char *pdata;
} RayCastTile;

typedef struct {
	float x, y, z;
	float dx, dy, dz;
	float r, g, b;
	short type;
} ThreeDeeLight;

typedef struct {
	Vector_f posn;
	Vector_f dirn;
	Vector_f up;
	float viewangle;
} ThreeDeeCamera;

typedef struct {
	char name[100];
	short numverts, numnorms, numpolys;
	Vertex_f vert[MAX_VERTS_PER_MODEL];
	Pgon poly[MAX_POLYS_PER_MODEL];
	Vector_f normal[MAX_VERTS_PER_MODEL];
	short shade;							// flag for shading
	float maxdim;							// max dimension for inst clip
} ThreeDeeModel;

typedef struct {
	short numverts, numpolys;
	Vertex_f vert[MAX_VERTS_PER_LANDSCAPE];
	Pgon poly[MAX_POLYS_PER_LANDSCAPE];
	Vector_f normal[MAX_VERTS_PER_LANDSCAPE];
	Vector_f	centroid[MAX_VERTS_PER_LANDSCAPE];		// centroids for filtering
	short shade;							// flag for shading
	float maxdim;							// max dimension for inst clip
} ThreeDeeLandscape;

typedef struct {
	ThreeDeeModel *model;
	void *attach;							// to point to another ThreeDeeInst
	short flags;							// ADD_TO_SURFLIST etc
	short shade, active, visible, clipped;
	short preshade;
	rgb_f pscolour[30];						// inst preshade flag and preshaded poly colours
	Vector_f posn;
	Vector_f rotn;
	Vector_f scal;
	Vector_f dirn;							// dirn of motion
	float dist, maxdist;
	float speed, maxspeed, accel;			// speed of motion
	float turn, maxturn;					// manouvreability
	short energy;
	short behaviour, counter;				// behaviour stuff
    short nextbehaviour, nextcounter;		// behaviour stuff
	float tx, ty, tz;						// behaviour stuff
	short effect;							// special effect for instance
} ThreeDeeInstance;

typedef struct {
	int numlights;
	int nummodels;
	int numinsts;
	int numtex;
	ThreeDeeLight light[3];
	ThreeDeeLight ambient;
	ThreeDeeCamera camera;
	ThreeDeeModel *pmodel[MAX_MODELS];			// pointers to models
	ThreeDeeInstance *pinst[MAX_INSTS];			// pointers to instances
	ThreeDeeInstance *pland;						// pntr to landscape inst
	ThreeDeeLandscape *plandmodel;				// ptr to landscape model
	char texname[16][80];
	BITMAP *ptex[16];									// pointers to texmaps
	rgb_f fogcolour;
	float fogdistance;					// if 0, fog is off
	BITMAP *background;					// background bitmap
	rgb_f backcolour;						// background colour if no bitmap
	float renderdist;
	char modelname[MAX_MODELS][40];
	RayCastTile tile;
} ThreeDeeScene;

ThreeDeeScene E3D_scene;

//
// ------------------- FUNCTION DEFINITIONS -----------------------
//
// initialise 3D engine
// render window size = w * h
int E3D_Init(int w, int h, int num_models, int num_textures, int num_insts);

// camera handling functions
void E3D_SetCameraPosn(float x, float y, float z);
void E3D_SetCameraRotn(float heading, float pitch, float roll);

// light handling functions
void E3D_SetLightDirn(short n, float r, float g, float b);
void E3D_SetLightColour(short n, float r, float g, float b);

// fog handling functions
void E3D_SetFogColour(float r, float g, float b);
void E3D_SetFogDistance(float dist);	// default = 0.0 = "off"

void E3D_SetBackColour(float r, float g, float b);

// Ray-cast ground
int E3D_SetGround(BITMAP *tilebmp);
int E3D_RenderGround(void);

// Instance handling
void E3D_RotateInst(ThreeDeeInstance *pinst, float rx, float ry, float rz);
void E3D_ScaleInst(ThreeDeeInstance *pinst, float sx, float sy, float sz);
void E3D_XlateInst(ThreeDeeInstance *pinst, float x, float y, float z);
int E3D_DestroyInst(ThreeDeeInstance *pinst);

// render the scene
int E3D_DrawScene(int x_offset, int y_offset, int vblsync);

// file i/o functions
int E3D_LoadScene(char *filename);
void E3D_ResetModel(ThreeDeeModel *model);
int E3D_Load3DO(char *filename, ThreeDeeModel *model);
int E3D_Save3DO(char *filename, ThreeDeeModel *model);
int E3D_LoadTexture(char *filename, BITMAP *ptex);
int E3D_LoadTextureList(char *filename, ThreeDeeScene *scene);
void E3D_CalculateNormals(ThreeDeeModel *model);

// instance handling functions
ThreeDeeInstance *E3D_CreateInst(ThreeDeeModel *model);

#define FOV		48.0
